
float4x4 oblv_ProjectionTransform_MAINPASS;

#define	farZ	oblv_ProjectionDepthRange_MAINPASS.x
#define	nearZ	oblv_ProjectionDepthRange_MAINPASS.y
#define	rangeZ	oblv_ProjectionDepthRange_MAINPASS.z
#define	sumZ	oblv_ProjectionDepthRange_MAINPASS.w
// .x - nearZ
// .y - farZ
// .z - farZ - nearZ
// .w - farZ + nearZ
float4 oblv_ProjectionDepthRange_MAINPASS;

texture zbufferTexture;
sampler zbufferSampler = sampler_state
{
	texture = <zbufferTexture>;

	AddressU = CLAMP;
	AddressV = CLAMP;

	MINFILTER = POINT;
	MAGFILTER = POINT;
};

// http://www.gamasutra.com/blogs/BranoKemen/20090812/2725/Logarithmic_Depth_Buffer.php
// http://mynameismjp.wordpress.com/2009/05/05/reconstructing-position-from-depth-continued/

/* ----------------------------------------------------------------------------- */

float radFOV(float nearL) {
#define	M_PI_2	1.57079632679489661923
#define	acot(x)	(M_PI_2 - atan(x))

	return acot(nearL) * 1;
}

float radFOVh() {
	/* horizontal FOV */
	return radFOV(oblv_ProjectionTransform_MAINPASS[0][0]);
}

float radFOVv() {
	/* vertical FOV */
	return radFOV(oblv_ProjectionTransform_MAINPASS[1][1]);
}

/* ----------------------------------------------------------------------------- */

float ReadZ(float2 pos) {
#if	(IN_RAWZ == 1)
	/* convert from RAWZ to float */
	float3 rawz = tex2D(zbufferSampler, pos).arg;
	float3 rawv = floor(255.0 * rawz + 0.5);

	float z = dot(rawv, float3(
		0.996093809371817670572857294849,
		0.0038909914428586627756752238080039,
		1.5199185323666651467481343000015e-5
	) / 255.0);
#else
	/* just read INTZ|DF?? */
	float z = tex2D(zbufferSampler, pos).x;
#endif

	return z;
}

float ReadL(float2 pos) {
	/* just read INTZ|DF?? */
	float z = tex2D(zbufferSampler, pos).x;

	return z;
}

/* ----------------------------------------------------------------------------- */

//  z_buffer_value = (1<<N) * ( a + b / z )
//
//  Where:
//
//     N = number of bits of Z precision
//     a = zFar / ( zFar - zNear )
//     b = zFar * zNear / ( zNear - zFar )
//     b = zFar * -zNear / ( zFar - zNear )
//     z = distance from the eye to the object
//
//  ...and z_buffer_value is an integer.
//
//  b / (z_buffer_value - a) = z

//static float ProjectionA =       farZ      / rangeZ;
//static float ProjectionB = (-farZ * nearZ) / rangeZ;
#define	ProjectionA	oblv_ProjectionTransform_MAINPASS[2][2]
#define	ProjectionB	oblv_ProjectionTransform_MAINPASS[3][2]

float NonLinearDepth(float posZ) {
	// Converts z-buffer depth to linear view-space depth (original range)
	return ((ProjectionB / posZ) + ProjectionA);
}

float LinearDepth(float posZ) {
	// Converts z-buffer depth to linear view-space depth (original range)
	return (ProjectionB / (posZ - ProjectionA));
}

/* . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . */

float HawkleyFoxDepth(float posZ) {
	// The HawkleyFox depth linearisation method:
	float depth = pow(abs(posZ), 0.05);

	return (2.0f * nearZ) / (sumZ - depth * rangeZ);
}

/* ----------------------------------------------------------------------------- */

#if 0
float4 Normals(VSOUT IN) : COLOR0
{
	float depth = readDepth(IN.UVCoord);
	float3 pos = getPosition(IN.UVCoord, depth);

	/* TODO: use ddx/ddy */
	float3 left  = pos - getPosition(IN.UVCoord + rcpres * float2(-1, 0), readDepth(IN.UVCoord + rcpres * float2(-1, 0)));
	float3 right =       getPosition(IN.UVCoord + rcpres * float2( 1, 0), readDepth(IN.UVCoord + rcpres * float2(1, 0) ) ) - pos;
	float3 up    = pos - getPosition(IN.UVCoord + rcpres * float2(0, -1), readDepth(IN.UVCoord + rcpres * float2(0, -1)) );
	float3 down  =       getPosition(IN.UVCoord + rcpres * float2(0,  1), readDepth(IN.UVCoord + rcpres * float2(0, 1))) - pos;

	float3 dx = length(left) < length(right) ? left : right;
	float3 dy = length(up) < length(down) ? up : down;

	float3 norm = normalize(cross(dx,dy));
	norm.z *= -1;

	return float4((norm+1)/2,1);
}
#endif

/* ----------------------------------------------------------------------------- */

struct VSIN
{
	float4 vertPos : POSITION0;
	float2 UVCoord : TEXCOORD0;
};

struct VSOUTc
{
	float4 vertPos : POSITION;
	float2 UVCoord : TEXCOORD0;
};

struct VSOUTp
{
	float4 vertPos : POSITION;
	float2 UVCoord : TEXCOORD0;
	float2 vertPrj : TEXCOORD1;
	float2 projMck : TEXCOORD2;
};

VSOUTc CopyVS(VSIN IN)
{
	VSOUTc OUT = (VSOUTc)0.0f;	// initialize to zero, avoid complaints.

	OUT.vertPos = IN.vertPos;
	OUT.UVCoord = IN.UVCoord;

	return OUT;
}

VSOUTc LinearizeVS(VSIN IN)
{
	VSOUTc OUT = (VSOUTc)0.0f;	// initialize to zero, avoid complaints.

	OUT.vertPos = IN.vertPos;
	OUT.UVCoord = IN.UVCoord;

	return OUT;
}

VSOUTc NonLinearizeVS(VSIN IN)
{
	VSOUTc OUT = (VSOUTc)0.0f;	// initialize to zero, avoid complaints.

	OUT.vertPos = IN.vertPos;
	OUT.UVCoord = IN.UVCoord;

	return OUT;
}

VSOUTp ProjectVS(VSIN IN)
{
	VSOUTp OUT = (VSOUTp)0.0f;	// initialize to zero, avoid complaints.

	OUT.vertPos = IN.vertPos;
	OUT.UVCoord = IN.UVCoord;
	OUT.vertPrj = IN.vertPos.xy;
	OUT.projMck = -1.0 / float2(
		oblv_ProjectionTransform_MAINPASS[0][0],
		oblv_ProjectionTransform_MAINPASS[1][1]
	);

	return OUT;
}

float4 CopyPS(VSOUTc IN) : COLOR0
{
	float d = ReadZ(IN.UVCoord);

	return float4(d, 0, 0, 0);
}

float4 LinearizePS(VSOUTc IN) : COLOR0
{
	float d = ReadZ(IN.UVCoord);

	/* [0,1] */
	float z = LinearDepth(d) / -rangeZ;
//	float z = HawkleyFoxDepth(z);

	return float4(z, 0, 0, 0);
}

float4 NonLinearizePS(VSOUTc IN) : COLOR0
{
	float d = ReadL(IN.UVCoord);

	/* [0,1] */
	float z = NonLinearDepth(d * -rangeZ);

	return float4(z, 0, 0, 0);
}

float4 ProjectPS(VSOUTp IN) : COLOR0
{
	float d = ReadZ(IN.UVCoord);

	/* [near,far] */
	float z = LinearDepth(d) / -rangeZ;
//	float z = HawkleyFoxDepth(d);

	/* [wx,wy] */
	float2 xy = IN.vertPrj * IN.projMck * d / -rangeZ;

	return float4(z, xy, d);
}

technique copy
{
	pass p0
	{
		VertexShader = compile vs_3_0 CopyVS();
		PixelShader  = compile ps_3_0 CopyPS();
	}
}

technique linearize
{
	pass p0
	{
		VertexShader = compile vs_3_0 LinearizeVS();
		PixelShader  = compile ps_3_0 LinearizePS();
	}
}

technique nonlinearize
{
	pass p0
	{
		VertexShader = compile vs_3_0 NonLinearizeVS();
		PixelShader  = compile ps_3_0 NonLinearizePS();
	}
}

technique project
{
	pass p0
	{
		VertexShader = compile vs_3_0 ProjectVS();
		PixelShader  = compile ps_3_0 ProjectPS();
	}
}

#if 0
technique normal
{
	pass p0
	{
		VertexShader = compile vs_3_0 ProjectVS();
		PixelShader  = compile ps_3_0 ProjectPS();
	}

	pass p1
	{
		VertexShader = compile vs_3_0 NormalVS();
		PixelShader  = compile ps_3_0 NormalPS();
	}
}
#endif
